from __future__ import annotations

from contextvars import ContextVar

##


_GLOBAL_BREAKPOINT = ContextVar("GLOBAL_BREAKPOINT", default=False)


def global_breakpoint() -> None:
    """Set a breakpoint if the global breakpoint is enabled."""
    if _GLOBAL_BREAKPOINT.get():  # pragma: no cover
        breakpoint()  # noqa: T100


def set_global_breakpoint() -> None:
    """Set the global breakpoint ."""
    _ = _GLOBAL_BREAKPOINT.set(True)


__all__ = ["global_breakpoint", "set_global_breakpoint"]



from dataclasses import dataclass
from typing import Literal

TrueOrFalseNoFutureLit = Literal["true", "false"]
type TrueOrFalseNoFutureTypeLit = Literal["true", "false"]


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class DataClassNoFutureInt:
    int_: int


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class DataClassNoFutureIntDefault:
    int_: int = 0


@dataclass(kw_only=True)
class DataClassNoFutureNestedInnerFirstInner:
    int_: int


@dataclass(kw_only=True)
class DataClassNoFutureNestedInnerFirstOuter:
    inner: DataClassNoFutureNestedInnerFirstInner


@dataclass(kw_only=True)
class DataClassNoFutureNestedOuterFirstOuter:
    inner: "DataClassNoFutureNestedOuterFirstInner"


@dataclass(kw_only=True)
class DataClassNoFutureNestedOuterFirstInner:
    int_: int



from __future__ import annotations

from dataclasses import dataclass
from re import IGNORECASE, search
from typing import Any, override


class _Meta(type):
    """Metaclass for the sentinel."""

    instance: Any = None

    @override
    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)
        return cls.instance


SENTINEL_REPR = "<sentinel>"


class Sentinel(metaclass=_Meta):
    """Base class for the sentinel object."""

    @override
    def __repr__(self) -> str:
        return SENTINEL_REPR

    @override
    def __str__(self) -> str:
        return repr(self)


sentinel = Sentinel()


##


def parse_sentinel(text: str, /) -> Sentinel:
    """Parse text into the Sentinel value."""
    if search("^(|sentinel|<sentinel>)$", text, flags=IGNORECASE):
        return sentinel
    raise ParseSentinelError(text=text)


@dataclass(kw_only=True, slots=True)
class ParseSentinelError(Exception):
    text: str

    @override
    def __str__(self) -> str:
        return f"Unable to parse sentinel value; got {self.text!r}"


__all__ = [
    "SENTINEL_REPR",
    "ParseSentinelError",
    "Sentinel",
    "parse_sentinel",
    "sentinel",
]



from __future__ import annotations

import re
from dataclasses import dataclass, field
from functools import cached_property
from logging import (
    FileHandler,
    Formatter,
    Handler,
    Logger,
    LogRecord,
    StreamHandler,
    basicConfig,
    getLevelNamesMapping,
    getLogger,
    setLogRecordFactory,
)
from logging.handlers import BaseRotatingHandler, TimedRotatingFileHandler
from pathlib import Path
from re import Pattern
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    NotRequired,
    Self,
    TypedDict,
    assert_never,
    cast,
    override,
)

from whenever import PlainDateTime, ZonedDateTime

from utilities.atomicwrites import move_many
from utilities.dataclasses import replace_non_sentinel
from utilities.errors import ImpossibleCaseError
from utilities.iterables import OneEmptyError, always_iterable, one
from utilities.pathlib import ensure_suffix, get_path
from utilities.re import (
    ExtractGroupError,
    ExtractGroupsError,
    extract_group,
    extract_groups,
)
from utilities.sentinel import Sentinel, sentinel
from utilities.tzlocal import LOCAL_TIME_ZONE_NAME
from utilities.whenever import (
    WheneverLogRecord,
    format_compact,
    get_now_local,
    to_local_plain,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping
    from datetime import time
    from logging import _FilterType

    from utilities.types import (
        LoggerOrName,
        LogLevel,
        MaybeCallablePathLike,
        MaybeIterable,
        PathLike,
    )


_DEFAULT_FORMAT = (
    "{zoned_datetime} | {name}:{funcName}:{lineno} | {levelname:8} | {message}"
)
_DEFAULT_DATEFMT = "%Y-%m-%d %H:%M:%S"
_DEFAULT_BACKUP_COUNT: int = 100
_DEFAULT_MAX_BYTES: int = 10 * 1024**2
_DEFAULT_WHEN: _When = "D"


##


def add_filters(handler: Handler, /, *filters: _FilterType) -> None:
    """Add a set of filters to a handler."""
    for filter_i in filters:
        handler.addFilter(filter_i)


##


def basic_config(
    *,
    obj: LoggerOrName | Handler | None = None,
    format_: str = _DEFAULT_FORMAT,
    datefmt: str = _DEFAULT_DATEFMT,
    level: LogLevel = "INFO",
    filters: MaybeIterable[_FilterType] | None = None,
    plain: bool = False,
    color_field_styles: Mapping[str, _FieldStyleKeys] | None = None,
) -> None:
    """Do the basic config."""
    match obj:
        case None:
            basicConfig(format=format_, datefmt=datefmt, style="{", level=level)
        case Logger() as logger:
            logger.setLevel(level)
            logger.addHandler(handler := StreamHandler())
            basic_config(
                obj=handler,
                format_=format_,
                datefmt=datefmt,
                level=level,
                filters=filters,
                plain=plain,
                color_field_styles=color_field_styles,
            )
        case str() as name:
            basic_config(
                obj=get_logger(logger=name),
                format_=format_,
                datefmt=datefmt,
                level=level,
                filters=filters,
                plain=plain,
                color_field_styles=color_field_styles,
            )
        case Handler() as handler:
            handler.setLevel(level)
            if filters is not None:
                add_filters(handler, *always_iterable(filters))
            formatter = get_formatter(
                format_=format_,
                datefmt=datefmt,
                plain=plain,
                color_field_styles=color_field_styles,
            )
            handler.setFormatter(formatter)
        case _ as never:
            assert_never(never)


##


def filter_for_key(
    key: str, /, *, default: bool = False
) -> Callable[[LogRecord], bool]:
    """Make a filter for a given attribute."""
    if (key in _FILTER_FOR_KEY_BLACKLIST) or key.startswith("__"):
        raise FilterForKeyError(key=key)

    def filter_(record: LogRecord, /) -> bool:
        try:
            value = getattr(record, key)
        except AttributeError:
            return default
        return bool(value)

    return filter_


# fmt: off
_FILTER_FOR_KEY_BLACKLIST = {
    "args", "created", "exc_info", "exc_text", "filename", "funcName", "getMessage", "levelname", "levelno", "lineno", "module", "msecs", "msg", "name", "pathname", "process", "processName", "relativeCreated", "stack_info", "taskName", "thread", "threadName"
}
# fmt: on


@dataclass(kw_only=True, slots=True)
class FilterForKeyError(Exception):
    key: str

    @override
    def __str__(self) -> str:
        return f"Invalid key: {self.key!r}"


##


type _FieldStyleKeys = Literal[
    "asctime", "hostname", "levelname", "name", "programname", "username"
]


class _FieldStyleDict(TypedDict):
    color: str
    bold: NotRequired[bool]


def get_formatter(
    *,
    format_: str = _DEFAULT_FORMAT,
    datefmt: str = _DEFAULT_DATEFMT,
    plain: bool = False,
    color_field_styles: Mapping[str, _FieldStyleKeys] | None = None,
) -> Formatter:
    """Get the formatter; colored if available."""
    setLogRecordFactory(WheneverLogRecord)
    if plain:
        return _get_plain_formatter(format_=format_, datefmt=datefmt)
    try:
        from coloredlogs import DEFAULT_FIELD_STYLES, ColoredFormatter
    except ModuleNotFoundError:  # pragma: no cover
        return _get_plain_formatter(format_=format_, datefmt=datefmt)
    default = cast("dict[_FieldStyleKeys, _FieldStyleDict]", DEFAULT_FIELD_STYLES)
    field_styles = {cast("str", k): v for k, v in default.items()}
    field_styles["zoned_datetime"] = default["asctime"]
    if color_field_styles is not None:
        field_styles.update({k: default[v] for k, v in color_field_styles.items()})
    return ColoredFormatter(
        fmt=format_, datefmt=datefmt, style="{", field_styles=field_styles
    )


def _get_plain_formatter(
    *, format_: str = _DEFAULT_FORMAT, datefmt: str = _DEFAULT_DATEFMT
) -> Formatter:
    """Get the plain formatter."""
    return Formatter(fmt=format_, datefmt=datefmt, style="{")


##


def get_logger(*, logger: LoggerOrName | None = None) -> Logger:
    """Get a logger."""
    match logger:
        case Logger():
            return logger
        case str() | None:
            return getLogger(logger)
        case _ as never:
            assert_never(never)


##


def get_logging_level_number(level: LogLevel, /) -> int:
    """Get the logging level number."""
    mapping = getLevelNamesMapping()
    try:
        return mapping[level]
    except KeyError:
        raise GetLoggingLevelNumberError(level=level) from None


@dataclass(kw_only=True, slots=True)
class GetLoggingLevelNumberError(Exception):
    level: LogLevel

    @override
    def __str__(self) -> str:
        return f"Invalid logging level: {self.level!r}"


##


def setup_logging(
    *,
    logger: LoggerOrName | None = None,
    format_: str = _DEFAULT_FORMAT,
    datefmt: str = _DEFAULT_DATEFMT,
    console_level: LogLevel = "INFO",
    console_prefix: str = "❯",  # noqa: RUF001
    console_filters: MaybeIterable[_FilterType] | None = None,
    files_dir: MaybeCallablePathLike | None = None,
    files_max_bytes: int = _DEFAULT_MAX_BYTES,
    files_when: _When = _DEFAULT_WHEN,
    files_interval: int = 1,
    files_backup_count: int = _DEFAULT_BACKUP_COUNT,
    files_filters: Iterable[_FilterType] | None = None,
) -> None:
    """Set up logger."""
    basic_config(
        obj=logger,
        format_=f"{console_prefix} {format_}",
        datefmt=datefmt,
        level=console_level,
        filters=console_filters,
    )
    logger_use = get_logger(logger=logger)
    name = logger_use.name
    dir_ = get_path(path=files_dir)
    levels: list[LogLevel] = ["DEBUG", "INFO", "ERROR"]
    for level in levels:
        lower = level.lower()
        for stem in [lower, f"{name}-{lower}"]:
            handler = SizeAndTimeRotatingFileHandler(
                dir_.joinpath(stem).with_suffix(".txt"),
                maxBytes=files_max_bytes,
                when=files_when,
                interval=files_interval,
                backupCount=files_backup_count,
            )
            logger_use.addHandler(handler)
            basic_config(
                obj=handler,
                format_=format_,
                datefmt=datefmt,
                level=level,
                filters=files_filters,
                plain=True,
            )


##


type _When = Literal[
    "S", "M", "H", "D", "midnight", "W0", "W1", "W2", "W3", "W4", "W5", "W6"
]


class SizeAndTimeRotatingFileHandler(BaseRotatingHandler):
    """Handler which rotates on size & time."""

    stream: Any

    @override
    def __init__(
        self,
        filename: PathLike,
        mode: Literal["a", "w", "x"] = "a",
        encoding: str | None = None,
        delay: bool = False,
        errors: Literal["strict", "ignore", "replace"] | None = None,
        maxBytes: int = _DEFAULT_MAX_BYTES,
        when: _When = _DEFAULT_WHEN,
        interval: int = 1,
        backupCount: int = _DEFAULT_BACKUP_COUNT,
        utc: bool = False,
        atTime: time | None = None,
    ) -> None:
        path = Path(filename)
        path.parent.mkdir(parents=True, exist_ok=True)
        super().__init__(path, mode, encoding=encoding, delay=delay, errors=errors)
        self._max_bytes = maxBytes if maxBytes >= 1 else None
        self._backup_count = backupCount if backupCount >= 1 else None
        self._filename = Path(self.baseFilename)
        self._directory = self._filename.parent
        self._stem = self._filename.stem
        self._suffix = self._filename.suffix
        self._patterns = _compute_rollover_patterns(self._stem, self._suffix)
        self._time_handler = TimedRotatingFileHandler(
            path,
            when=when,
            interval=interval,
            backupCount=backupCount,
            encoding=encoding,
            delay=delay,
            utc=utc,
            atTime=atTime,
            errors=errors,
        )

    @override
    def emit(self, record: LogRecord) -> None:
        try:
            if (self._backup_count is not None) and self._should_rollover(record):
                self._do_rollover(  # skipif-ci-and-windows
                    backup_count=self._backup_count
                )
            FileHandler.emit(self, record)
        except Exception:  # noqa: BLE001  # pragma: no cover
            self.handleError(record)

    def _do_rollover(self, *, backup_count: int = 1) -> None:
        if self.stream:  # pragma: no cover
            self.stream.close()
            self.stream = None

        actions = _compute_rollover_actions(  # skipif-ci-and-windows
            self._directory,
            self._stem,
            self._suffix,
            patterns=self._patterns,
            backup_count=backup_count,
        )
        actions.do()  # skipif-ci-and-windows

        if not self.delay:  # pragma: no cover
            self.stream = self._open()
        self._time_handler.rolloverAt = (  # skipif-ci-and-windows
            self._time_handler.computeRollover(get_now_local().timestamp())
        )

    def _should_rollover(self, record: LogRecord, /) -> bool:
        if self._max_bytes is not None:  # skipif-ci-and-windows
            try:
                size = self._filename.stat().st_size
            except FileNotFoundError:
                pass
            else:
                if size >= self._max_bytes:
                    return True
        return bool(self._time_handler.shouldRollover(record))  # skipif-ci-and-windows


def _compute_rollover_patterns(stem: str, suffix: str, /) -> _RolloverPatterns:
    return _RolloverPatterns(
        pattern1=re.compile(rf"^{stem}\.(\d+){suffix}$"),
        pattern2=re.compile(rf"^{stem}\.(\d+)__([\dT]+?){suffix}$"),
        pattern3=re.compile(rf"^{stem}\.(\d+)__([\dT]+?)__([\dT]+?){suffix}$"),
    )


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class _RolloverPatterns:
    pattern1: Pattern[str]
    pattern2: Pattern[str]
    pattern3: Pattern[str]


def _compute_rollover_actions(
    directory: Path,
    stem: str,
    suffix: str,
    /,
    *,
    patterns: _RolloverPatterns | None = None,
    backup_count: int = 1,
) -> _RolloverActions:
    patterns = (
        _compute_rollover_patterns(stem, suffix) if patterns is None else patterns
    )
    files = {
        file
        for path in directory.iterdir()
        if (file := _RotatingLogFile.from_path(path, stem, suffix, patterns=patterns))
        is not None
    }
    deletions: set[_Deletion] = set()
    rotations: set[_Rotation] = set()
    for file in files:
        match file.index, file.start, file.end:
            case int() as index, _, _ if index >= backup_count:
                deletions.add(_Deletion(file=file))
            case index, None, _:
                if index is None:
                    curr = 0
                    end = get_now_local()
                else:
                    curr = index
                    end = sentinel
                try:
                    start = one(f for f in files if f.index == curr + 1).end
                except OneEmptyError:
                    start = None
                rotations.add(
                    _Rotation(file=file, index=curr + 1, start=start, end=end)
                )
            case int() as index, ZonedDateTime(), ZonedDateTime():
                rotations.add(_Rotation(file=file, index=index + 1))
            case _:  # pragma: no cover
                raise NotImplementedError
    return _RolloverActions(deletions=deletions, rotations=rotations)


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class _RolloverActions:
    deletions: set[_Deletion] = field(default_factory=set)
    rotations: set[_Rotation] = field(default_factory=set)

    def do(self) -> None:
        for deletion in self.deletions:
            deletion.delete()
        move_many(*((r.file.path, r.destination) for r in self.rotations))


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class _RotatingLogFile:
    directory: Path
    stem: str
    suffix: str
    index: int | None = None
    start: ZonedDateTime | None = None
    end: ZonedDateTime | None = None

    @classmethod
    def from_path(
        cls,
        path: Path,
        stem: str,
        suffix: str,
        /,
        *,
        patterns: _RolloverPatterns | None = None,
    ) -> Self | None:
        if (not path.stem.startswith(stem)) or path.suffix != suffix:
            return None
        if patterns is None:
            patterns = _compute_rollover_patterns(stem, suffix)
        try:
            index, start, end = extract_groups(patterns.pattern3, path.name)
        except ExtractGroupsError:
            pass
        else:
            return cls(
                directory=path.parent,
                stem=stem,
                suffix=suffix,
                index=int(index),
                start=PlainDateTime.parse_common_iso(start).assume_tz(
                    LOCAL_TIME_ZONE_NAME
                ),
                end=PlainDateTime.parse_common_iso(end).assume_tz(LOCAL_TIME_ZONE_NAME),
            )
        try:
            index, end = extract_groups(patterns.pattern2, path.name)
        except ExtractGroupsError:
            pass
        else:
            return cls(
                directory=path.parent,
                stem=stem,
                suffix=suffix,
                index=int(index),
                end=PlainDateTime.parse_common_iso(end).assume_tz(LOCAL_TIME_ZONE_NAME),
            )
        try:
            index = extract_group(patterns.pattern1, path.name)
        except ExtractGroupError:
            pass
        else:
            return cls(
                directory=path.parent, stem=stem, suffix=suffix, index=int(index)
            )
        return cls(directory=path.parent, stem=stem, suffix=suffix)

    @cached_property
    def path(self) -> Path:
        """The full path."""
        match self.index, self.start, self.end:
            case None, None, None:
                tail = None
            case int() as index, None, None:
                tail = str(index)
            case int() as index, None, ZonedDateTime() as end:
                tail = f"{index}__{format_compact(to_local_plain(end))}"
            case int() as index, ZonedDateTime() as start, ZonedDateTime() as end:
                tail = f"{index}__{format_compact(to_local_plain(start))}__{format_compact(to_local_plain(end))}"
            case _:  # pragma: no cover
                raise ImpossibleCaseError(
                    case=[f"{self.index=}", f"{self.start=}", f"{self.end=}"]
                )
        stem = self.stem if tail is None else f"{self.stem}.{tail}"
        return ensure_suffix(self.directory.joinpath(stem), self.suffix)

    def replace(
        self,
        *,
        index: int | None | Sentinel = sentinel,
        start: ZonedDateTime | None | Sentinel = sentinel,
        end: ZonedDateTime | None | Sentinel = sentinel,
    ) -> Self:
        return replace_non_sentinel(self, index=index, start=start, end=end)


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class _Deletion:
    file: _RotatingLogFile

    def delete(self) -> None:
        self.file.path.unlink(missing_ok=True)


@dataclass(order=True, unsafe_hash=True, kw_only=True)
class _Rotation:
    file: _RotatingLogFile
    index: int = 0
    start: ZonedDateTime | None | Sentinel = sentinel
    end: ZonedDateTime | Sentinel = sentinel

    @cached_property
    def destination(self) -> Path:
        return self.file.replace(index=self.index, start=self.start, end=self.end).path


__all__ = [
    "FilterForKeyError",
    "GetLoggingLevelNumberError",
    "SizeAndTimeRotatingFileHandler",
    "add_filters",
    "basic_config",
    "filter_for_key",
    "get_logger",
    "get_logging_level_number",
    "setup_logging",
]



from __future__ import annotations

from re import search
from typing import TYPE_CHECKING

from hypothesis import given
from hypothesis.strategies import booleans, integers, none
from pytest import raises

from utilities.hypothesis import sentinels, text_ascii, versions
from utilities.version import (
    ParseVersionError,
    Version,
    _VersionEmptySuffixError,
    _VersionNegativeMajorVersionError,
    _VersionNegativeMinorVersionError,
    _VersionNegativePatchVersionError,
    _VersionZeroError,
    get_version,
    parse_version,
)

if TYPE_CHECKING:
    from utilities.sentinel import Sentinel


class TestGetVersion:
    @given(version=versions())
    def test_version(self, *, version: Version) -> None:
        assert get_version(version=version) == version

    @given(version=versions())
    def test_str(self, *, version: Version) -> None:
        assert get_version(version=str(version)) == version

    @given(version=none() | sentinels())
    def test_none_or_sentinel(self, *, version: None | Sentinel) -> None:
        assert get_version(version=version) is version

    @given(version=versions())
    def test_callable(self, *, version: Version) -> None:
        assert get_version(version=lambda: version) == version


class TestParseVersion:
    @given(version=versions())
    def test_main(self, *, version: Version) -> None:
        parsed = parse_version(str(version))
        assert parsed == version

    def test_error(self) -> None:
        with raises(ParseVersionError, match="Invalid version string: 'invalid'"):
            _ = parse_version("invalid")


class TestVersion:
    @given(version=versions())
    def test_hashable(self, *, version: Version) -> None:
        _ = hash(version)

    @given(version1=versions(), version2=versions())
    def test_orderable(self, *, version1: Version, version2: Version) -> None:
        assert (version1 <= version2) or (version1 >= version2)

    @given(version=versions(suffix=booleans()))
    def test_repr(self, *, version: Version) -> None:
        result = repr(version)
        assert search(r"^\d+\.\d+\.\d+", result)

    @given(version=versions())
    def test_bump_major(self, *, version: Version) -> None:
        bumped = version.bump_major()
        assert version < bumped
        assert bumped.major == version.major + 1
        assert bumped.minor == 0
        assert bumped.patch == 0
        assert bumped.suffix is None

    @given(version=versions())
    def test_bump_minor(self, *, version: Version) -> None:
        bumped = version.bump_minor()
        assert version < bumped
        assert bumped.major == version.major
        assert bumped.minor == version.minor + 1
        assert bumped.patch == 0
        assert bumped.suffix is None

    @given(version=versions())
    def test_bump_patch(self, *, version: Version) -> None:
        bumped = version.bump_patch()
        assert version < bumped
        assert bumped.major == version.major
        assert bumped.minor == version.minor
        assert bumped.patch == version.patch + 1
        assert bumped.suffix is None

    @given(version=versions(), suffix=text_ascii(min_size=1) | none())
    def test_with_suffix(self, *, version: Version, suffix: str | None) -> None:
        new = version.with_suffix(suffix=suffix)
        assert new.major == version.major
        assert new.minor == version.minor
        assert new.patch == version.patch
        assert new.suffix == suffix

    @given(version=versions())
    def test_error_order(self, *, version: Version) -> None:
        with raises(TypeError):
            _ = version <= None

    def test_error_zero(self) -> None:
        with raises(
            _VersionZeroError, match="Version must be greater than zero; got 0.0.0"
        ):
            _ = Version(0, 0, 0)

    @given(major=integers(max_value=-1))
    def test_error_negative_major_version(self, *, major: int) -> None:
        with raises(
            _VersionNegativeMajorVersionError,
            match="Major version must be non-negative; got .*",
        ):
            _ = Version(major=major)

    @given(minor=integers(max_value=-1))
    def test_error_negative_minor_version(self, *, minor: int) -> None:
        with raises(
            _VersionNegativeMinorVersionError,
            match="Minor version must be non-negative; got .*",
        ):
            _ = Version(minor=minor)

    @given(patch=integers(max_value=-1))
    def test_error_negative_patch_version(self, *, patch: int) -> None:
        with raises(
            _VersionNegativePatchVersionError,
            match="Patch version must be non-negative; got .*",
        ):
            _ = Version(patch=patch)

    def test_error_empty_suffix(self) -> None:
        with raises(_VersionEmptySuffixError, match="Suffix must be non-empty; got .*"):
            _ = Version(suffix="")



from __future__ import annotations

from utilities.asyncio import sleep_td
from utilities.atools import call_memoized
from utilities.whenever import SECOND


class TestCallMemoized:
    async def test_main(self) -> None:
        counter = 0

        async def increment() -> int:
            nonlocal counter
            counter += 1
            return counter

        for i in range(1, 3):
            assert (await call_memoized(increment)) == i
            assert counter == i

    async def test_refresh(self) -> None:
        counter = 0
        delta = 0.1 * SECOND

        async def increment() -> int:
            nonlocal counter
            counter += 1
            return counter

        for _ in range(2):
            assert (await call_memoized(increment, delta)) == 1
            assert counter == 1
        await sleep_td(2 * delta)
        for _ in range(2):
            assert (await call_memoized(increment, delta)) == 2
            assert counter == 2



from __future__ import annotations

import asyncio
from asyncio import (
    Event,
    Lock,
    PriorityQueue,
    Queue,
    QueueEmpty,
    QueueFull,
    Semaphore,
    StreamReader,
    Task,
    TaskGroup,
    create_subprocess_shell,
    sleep,
)
from collections.abc import (
    Callable,
    Hashable,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    ValuesView,
)
from contextlib import (
    AbstractAsyncContextManager,
    AsyncExitStack,
    _AsyncGeneratorContextManager,
    asynccontextmanager,
    suppress,
)
from dataclasses import dataclass
from io import StringIO
from itertools import chain
from subprocess import PIPE
from sys import stderr, stdout
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Self,
    TextIO,
    assert_never,
    overload,
    override,
)

from typing_extensions import deprecated

from utilities.functions import ensure_int, ensure_not_none, to_bool
from utilities.random import SYSTEM_RANDOM
from utilities.sentinel import Sentinel, sentinel
from utilities.types import (
    Delta,
    MaybeCallableBool,
    SupportsKeysAndGetItem,
    SupportsRichComparison,
)
from utilities.whenever import get_now, round_date_or_date_time, to_nanoseconds

if TYPE_CHECKING:
    from asyncio import _CoroutineLike
    from asyncio.subprocess import Process
    from collections import deque
    from collections.abc import AsyncIterator, Sequence
    from contextvars import Context
    from random import Random
    from types import TracebackType

    from whenever import ZonedDateTime

    from utilities.types import MaybeCallableEvent, MaybeType


class AsyncDict[K, V]:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, map: SupportsKeysAndGetItem[K, V], /) -> None: ...
    @overload
    def __init__(self, iterable: Iterable[tuple[K, V]], /) -> None: ...
    @override
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__()
        self._dict = dict[K, V](*args, **kwargs)
        self._lock = Lock()

    async def __aenter__(self) -> dict[K, V]:
        await self._lock.__aenter__()
        return self._dict

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
        /,
    ) -> None:
        await self._lock.__aexit__(exc_type, exc, tb)

    def __contains__(self, key: Any, /) -> bool:
        return key in self._dict

    @override
    def __eq__(self, other: Any, /) -> bool:
        return self._dict == other

    __hash__: ClassVar[None] = None  # pyright: ignore[reportIncompatibleMethodOverride]

    def __getitem__(self, key: K, /) -> V:
        return self._dict[key]

    def __iter__(self) -> Iterator[K]:
        yield from self._dict

    def __len__(self) -> int:
        return len(self._dict)

    @override
    def __repr__(self) -> str:
        return repr(self._dict)

    def __reversed__(self) -> Iterator[K]:
        return reversed(self._dict)

    @override
    def __str__(self) -> str:
        return str(self._dict)

    @property
    def empty(self) -> bool:
        return len(self) == 0

    @classmethod
    @overload
    def fromkeys[T](
        cls, iterable: Iterable[T], value: None = None, /
    ) -> AsyncDict[T, Any | None]: ...
    @classmethod
    @overload
    def fromkeys[K2, V2](
        cls, iterable: Iterable[K2], value: V2, /
    ) -> AsyncDict[K2, V2]: ...
    @classmethod
    def fromkeys(
        cls, iterable: Iterable[Any], value: Any = None, /
    ) -> AsyncDict[Any, Any]:
        return cls(dict.fromkeys(iterable, value))

    async def clear(self) -> None:
        async with self._lock:
            self._dict.clear()

    def copy(self) -> Self:
        return type(self)(self._dict.items())

    async def del_(self, key: K, /) -> None:
        async with self._lock:
            del self._dict[key]

    @overload
    def get(self, key: K, default: None = None, /) -> V | None: ...
    @overload
    def get(self, key: K, default: V, /) -> V: ...
    @overload
    def get[V2](self, key: K, default: V2, /) -> V | V2: ...
    def get(self, key: K, default: Any = sentinel, /) -> Any:
        match default:
            case Sentinel():
                return self._dict.get(key)
            case _:
                return self._dict.get(key, default)

    def keys(self) -> KeysView[K]:
        return self._dict.keys()

    def items(self) -> ItemsView[K, V]:
        return self._dict.items()

    @overload
    async def pop(self, key: K, /) -> V: ...
    @overload
    async def pop(self, key: K, default: V, /) -> V: ...
    @overload
    async def pop[V2](self, key: K, default: V2, /) -> V | V2: ...
    async def pop(self, key: K, default: Any = sentinel, /) -> Any:
        async with self._lock:
            match default:
                case Sentinel():
                    return self._dict.pop(key)
                case _:
                    return self._dict.pop(key, default)

    async def popitem(self) -> tuple[K, V]:
        async with self._lock:
            return self._dict.popitem()

    async def set(self, key: K, value: V, /) -> None:
        async with self._lock:
            self._dict[key] = value

    async def setdefault(self, key: K, default: V, /) -> V:
        async with self._lock:
            return self._dict.setdefault(key, default)

    @overload
    async def update(self, m: SupportsKeysAndGetItem[K, V], /) -> None: ...
    @overload
    async def update(self, m: Iterable[tuple[K, V]], /) -> None: ...
    async def update(self, *args: Any, **kwargs: V) -> None:
        async with self._lock:
            self._dict.update(*args, **kwargs)

    def values(self) -> ValuesView[V]:
        return self._dict.values()


##


class EnhancedQueue[T](Queue[T]):
    """An asynchronous deque."""

    @override
    def __init__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize=maxsize)
        self._finished: Event
        self._getters: deque[Any]
        self._putters: deque[Any]
        self._queue: deque[T]
        self._unfinished_tasks: int

    @override
    @deprecated("Use `get_left`/`get_right` instead")
    async def get(self) -> T:
        raise RuntimeError  # pragma: no cover

    @override
    @deprecated("Use `get_left_nowait`/`get_right_nowait` instead")
    def get_nowait(self) -> T:
        raise RuntimeError  # pragma: no cover

    @override
    @deprecated("Use `put_left`/`put_right` instead")
    async def put(self, item: T) -> None:
        raise RuntimeError(item)  # pragma: no cover

    @override
    @deprecated("Use `put_left_nowait`/`put_right_nowait` instead")
    def put_nowait(self, item: T) -> None:
        raise RuntimeError(item)  # pragma: no cover

    # get all

    async def get_all(self, *, reverse: bool = False) -> Sequence[T]:
        """Remove and return all items from the queue."""
        first = await (self.get_right() if reverse else self.get_left())
        return list(chain([first], self.get_all_nowait(reverse=reverse)))

    def get_all_nowait(self, *, reverse: bool = False) -> Sequence[T]:
        """Remove and return all items from the queue without blocking."""
        items: Sequence[T] = []
        while True:
            try:
                items.append(
                    self.get_right_nowait() if reverse else self.get_left_nowait()
                )
            except QueueEmpty:
                return items

    # get left/right

    async def get_left(self) -> T:
        """Remove and return an item from the start of the queue."""
        return await self._get_left_or_right(self._get)

    async def get_right(self) -> T:
        """Remove and return an item from the end of the queue."""
        return await self._get_left_or_right(self._get_right)

    def get_left_nowait(self) -> T:
        """Remove and return an item from the start of the queue without blocking."""
        return self._get_left_or_right_nowait(self._get)

    def get_right_nowait(self) -> T:
        """Remove and return an item from the end of the queue without blocking."""
        return self._get_left_or_right_nowait(self._get_right)

    # put left/right

    async def put_left(self, *items: T) -> None:
        """Put items into the queue at the start."""
        return await self._put_left_or_right(self._put_left, *items)

    async def put_right(self, *items: T) -> None:
        """Put items into the queue at the end."""
        return await self._put_left_or_right(self._put, *items)

    def put_left_nowait(self, *items: T) -> None:
        """Put items into the queue at the start without blocking."""
        self._put_left_or_right_nowait(self._put_left, *items)

    def put_right_nowait(self, *items: T) -> None:
        """Put items into the queue at the end without blocking."""
        self._put_left_or_right_nowait(self._put, *items)

    # private

    def _put_left(self, item: T) -> None:
        self._queue.appendleft(item)

    def _get_right(self) -> T:
        return self._queue.pop()

    async def _get_left_or_right(self, getter_use: Callable[[], T], /) -> T:
        while self.empty():  # pragma: no cover
            getter = self._get_loop().create_future()  # pyright: ignore[reportAttributeAccessIssue]
            self._getters.append(getter)
            try:
                await getter
            except:
                getter.cancel()
                with suppress(ValueError):
                    self._getters.remove(getter)
                if not self.empty() and not getter.cancelled():
                    self._wakeup_next(self._getters)  # pyright: ignore[reportAttributeAccessIssue]
                raise
        return getter_use()

    def _get_left_or_right_nowait(self, getter: Callable[[], T], /) -> T:
        if self.empty():
            raise QueueEmpty
        item = getter()
        self._wakeup_next(self._putters)  # pyright: ignore[reportAttributeAccessIssue]
        return item

    async def _put_left_or_right(
        self, putter_use: Callable[[T], None], /, *items: T
    ) -> None:
        """Put an item into the queue."""
        for item in items:
            await self._put_left_or_right_one(putter_use, item)

    async def _put_left_or_right_one(
        self, putter_use: Callable[[T], None], item: T, /
    ) -> None:
        """Put an item into the queue."""
        while self.full():  # pragma: no cover
            putter = self._get_loop().create_future()  # pyright: ignore[reportAttributeAccessIssue]
            self._putters.append(putter)
            try:
                await putter
            except:
                putter.cancel()
                with suppress(ValueError):
                    self._putters.remove(putter)
                if not self.full() and not putter.cancelled():
                    self._wakeup_next(self._putters)  # pyright: ignore[reportAttributeAccessIssue]
                raise
        return putter_use(item)

    def _put_left_or_right_nowait(
        self, putter: Callable[[T], None], /, *items: T
    ) -> None:
        for item in items:
            self._put_left_or_right_nowait_one(putter, item)

    def _put_left_or_right_nowait_one(
        self, putter: Callable[[T], None], item: T, /
    ) -> None:
        if self.full():  # pragma: no cover
            raise QueueFull
        putter(item)
        self._unfinished_tasks += 1
        self._finished.clear()
        self._wakeup_next(self._getters)  # pyright: ignore[reportAttributeAccessIssue]


##


class EnhancedTaskGroup(TaskGroup):
    """Task group with enhanced features."""

    _max_tasks: int | None
    _semaphore: Semaphore | None
    _timeout: Delta | None
    _error: MaybeType[BaseException]
    _debug: MaybeCallableBool
    _stack: AsyncExitStack
    _timeout_cm: _AsyncGeneratorContextManager[None] | None

    @override
    def __init__(
        self,
        *,
        max_tasks: int | None = None,
        timeout: Delta | None = None,
        error: MaybeType[BaseException] = TimeoutError,
        debug: MaybeCallableBool = False,
    ) -> None:
        super().__init__()
        self._max_tasks = max_tasks
        if (max_tasks is None) or (max_tasks <= 0):
            self._semaphore = None
        else:
            self._semaphore = Semaphore(max_tasks)
        self._timeout = timeout
        self._error = error
        self._debug = debug
        self._stack = AsyncExitStack()
        self._timeout_cm = None

    @override
    async def __aenter__(self) -> Self:
        _ = await self._stack.__aenter__()
        return await super().__aenter__()

    @override
    async def __aexit__(
        self,
        et: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        _ = await self._stack.__aexit__(et, exc, tb)
        match self._is_debug():
            case True:
                with suppress(Exception):
                    _ = await super().__aexit__(et, exc, tb)
            case False:
                _ = await super().__aexit__(et, exc, tb)
            case _ as never:
                assert_never(never)

    @override
    def create_task[T](
        self,
        coro: _CoroutineLike[T],
        *,
        name: str | None = None,
        context: Context | None = None,
    ) -> Task[T]:
        if self._semaphore is None:
            coroutine = coro
        else:
            coroutine = self._wrap_with_semaphore(self._semaphore, coro)
        coroutine = self._wrap_with_timeout(coroutine)
        return super().create_task(coroutine, name=name, context=context)

    def create_task_context[T](self, cm: AbstractAsyncContextManager[T], /) -> Task[T]:
        """Have the TaskGroup start an asynchronous context manager."""
        _ = self._stack.push_async_callback(cm.__aexit__, None, None, None)
        return self.create_task(cm.__aenter__())

    async def run_or_create_many_tasks[**P, T](
        self,
        make_coro: Callable[P, _CoroutineLike[T]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T | Sequence[Task[T]]:
        match self._is_debug(), self._max_tasks:
            case (True, _) | (False, None):
                return await make_coro(*args, **kwargs)
            case False, int():
                return [
                    self.create_task(make_coro(*args, **kwargs))
                    for _ in range(self._max_tasks)
                ]
            case _ as never:
                assert_never(never)

    async def run_or_create_task[T](
        self,
        coro: _CoroutineLike[T],
        *,
        name: str | None = None,
        context: Context | None = None,
    ) -> T | Task[T]:
        match self._is_debug():
            case True:
                return await coro
            case False:
                return self.create_task(coro, name=name, context=context)
            case _ as never:
                assert_never(never)

    def _is_debug(self) -> bool:
        return to_bool(bool_=self._debug) or (
            (self._max_tasks is not None) and (self._max_tasks <= 0)
        )

    async def _wrap_with_semaphore[T](
        self, semaphore: Semaphore, coroutine: _CoroutineLike[T], /
    ) -> T:
        async with semaphore:
            return await coroutine

    async def _wrap_with_timeout[T](self, coroutine: _CoroutineLike[T], /) -> T:
        async with timeout_td(self._timeout, error=self._error):
            return await coroutine


##


class UniquePriorityQueue[T: SupportsRichComparison, U: Hashable](
    PriorityQueue[tuple[T, U]]
):
    """Priority queue with unique tasks."""

    @override
    def __init__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize)
        self._set: set[U] = set()

    @override
    def _get(self) -> tuple[T, U]:
        item = super()._get()
        _, value = item
        self._set.remove(value)
        return item

    @override
    def _put(self, item: tuple[T, U]) -> None:
        _, value = item
        if value not in self._set:
            super()._put(item)
            self._set.add(value)


class UniqueQueue[T: Hashable](Queue[T]):
    """Queue with unique tasks."""

    @override
    def __init__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize)
        self._set: set[T] = set()

    @override
    def _get(self) -> T:
        item = super()._get()
        self._set.remove(item)
        return item

    @override
    def _put(self, item: T) -> None:
        if item not in self._set:
            super()._put(item)
            self._set.add(item)


##


@overload
def get_event(*, event: MaybeCallableEvent) -> Event: ...
@overload
def get_event(*, event: None) -> None: ...
@overload
def get_event(*, event: Sentinel) -> Sentinel: ...
@overload
def get_event(*, event: MaybeCallableEvent | Sentinel) -> Event | Sentinel: ...
@overload
def get_event(
    *, event: MaybeCallableEvent | None | Sentinel = sentinel
) -> Event | None | Sentinel: ...
def get_event(
    *, event: MaybeCallableEvent | None | Sentinel = sentinel
) -> Event | None | Sentinel:
    """Get the event."""
    match event:
        case Event() | None | Sentinel():
            return event
        case Callable() as func:
            return get_event(event=func())
        case _ as never:
            assert_never(never)


##


async def get_items[T](queue: Queue[T], /, *, max_size: int | None = None) -> list[T]:
    """Get items from a queue; if empty then wait."""
    try:
        items = [await queue.get()]
    except RuntimeError as error:  # pragma: no cover
        from utilities.pytest import is_pytest

        if (not is_pytest()) or (error.args[0] != "Event loop is closed"):
            raise
        return []
    max_size_use = None if max_size is None else (max_size - 1)
    items.extend(get_items_nowait(queue, max_size=max_size_use))
    return items


def get_items_nowait[T](queue: Queue[T], /, *, max_size: int | None = None) -> list[T]:
    """Get items from a queue; no waiting."""
    items: list[T] = []
    if max_size is None:
        while True:
            try:
                items.append(queue.get_nowait())
            except QueueEmpty:
                break
    else:
        while len(items) < max_size:
            try:
                items.append(queue.get_nowait())
            except QueueEmpty:
                break
    return items


##


async def put_items[T](items: Iterable[T], queue: Queue[T], /) -> None:
    """Put items into a queue; if full then wait."""
    for item in items:
        await queue.put(item)


def put_items_nowait[T](items: Iterable[T], queue: Queue[T], /) -> None:
    """Put items into a queue; no waiting."""
    for item in items:
        queue.put_nowait(item)


##


async def sleep_max(
    sleep: Delta | None = None, /, *, random: Random = SYSTEM_RANDOM
) -> None:
    """Sleep which accepts deltas."""
    if sleep is None:
        return
    await asyncio.sleep(random.uniform(0.0, to_nanoseconds(sleep) / 1e9))


##


async def sleep_rounded(delta: Delta, /) -> None:
    """Sleep until a rounded time."""
    await sleep_until(round_date_or_date_time(get_now(), delta, mode="ceil"))


##


async def sleep_td(delta: Delta | None = None, /) -> None:
    """Sleep which accepts deltas."""
    if delta is None:
        return
    await sleep(to_nanoseconds(delta) / 1e9)


##


async def sleep_until(datetime: ZonedDateTime, /) -> None:
    """Sleep until a given time."""
    await sleep_td(datetime - get_now())


##


@dataclass(kw_only=True, slots=True)
class StreamCommandOutput:
    process: Process
    stdout: str
    stderr: str

    @property
    def return_code(self) -> int:
        return ensure_int(self.process.returncode)  # skipif-not-windows


async def stream_command(cmd: str, /) -> StreamCommandOutput:
    """Run a shell command asynchronously and stream its output in real time."""
    process = await create_subprocess_shell(  # skipif-not-windows
        cmd, stdout=PIPE, stderr=PIPE
    )
    proc_stdout = ensure_not_none(  # skipif-not-windows
        process.stdout, desc="process.stdout"
    )
    proc_stderr = ensure_not_none(  # skipif-not-windows
        process.stderr, desc="process.stderr"
    )
    ret_stdout = StringIO()  # skipif-not-windows
    ret_stderr = StringIO()  # skipif-not-windows
    async with TaskGroup() as tg:  # skipif-not-windows
        _ = tg.create_task(_stream_one(proc_stdout, stdout, ret_stdout))
        _ = tg.create_task(_stream_one(proc_stderr, stderr, ret_stderr))
    _ = await process.wait()  # skipif-not-windows
    return StreamCommandOutput(  # skipif-not-windows
        process=process, stdout=ret_stdout.getvalue(), stderr=ret_stderr.getvalue()
    )


async def _stream_one(
    input_: StreamReader, out_stream: TextIO, ret_stream: StringIO, /
) -> None:
    """Asynchronously read from a stream and write to the target output stream."""
    while True:  # skipif-not-windows
        line = await input_.readline()
        if not line:
            break
        decoded = line.decode()
        _ = out_stream.write(decoded)
        out_stream.flush()
        _ = ret_stream.write(decoded)


##


@asynccontextmanager
async def timeout_td(
    timeout: Delta | None = None, /, *, error: MaybeType[BaseException] = TimeoutError
) -> AsyncIterator[None]:
    """Timeout context manager which accepts deltas."""
    timeout_use = None if timeout is None else (to_nanoseconds(timeout) / 1e9)
    try:
        async with asyncio.timeout(timeout_use):
            yield
    except TimeoutError:
        raise error from None


__all__ = [
    "AsyncDict",
    "EnhancedQueue",
    "EnhancedTaskGroup",
    "StreamCommandOutput",
    "UniquePriorityQueue",
    "UniqueQueue",
    "get_event",
    "get_items",
    "get_items_nowait",
    "put_items",
    "put_items_nowait",
    "sleep_max",
    "sleep_rounded",
    "sleep_td",
    "sleep_until",
    "stream_command",
    "timeout_td",
]



from __future__ import annotations

import gzip
from pathlib import Path
from typing import TYPE_CHECKING

from utilities.atomicwrites import writer

if TYPE_CHECKING:
    from utilities.types import PathLike


def read_binary(path: PathLike, /, *, decompress: bool = False) -> bytes:
    """Read a byte string from disk."""
    path = Path(path)
    if decompress:
        with gzip.open(path) as gz:
            return gz.read()
    else:
        return path.read_bytes()


def write_binary(
    data: bytes, path: PathLike, /, *, compress: bool = False, overwrite: bool = False
) -> None:
    """Write a byte string to disk."""
    with writer(path, compress=compress, overwrite=overwrite) as temp:
        _ = temp.write_bytes(data)


__all__ = ["read_binary", "write_binary"]



from __future__ import annotations

from collections.abc import Callable, Hashable
from dataclasses import dataclass
from itertools import islice
from textwrap import indent
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeGuard,
    assert_never,
    cast,
    overload,
    override,
)

from more_itertools import bucket, partition, split_into
from more_itertools import peekable as _peekable

from utilities.functions import get_class_name
from utilities.iterables import OneNonUniqueError, one
from utilities.reprlib import get_repr
from utilities.sentinel import Sentinel, sentinel

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, Sequence


@overload
def bucket_mapping[T, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: None = None,
    post: None = None,
) -> Mapping[UH, Iterator[T]]: ...
@overload
def bucket_mapping[T, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: None = None,
    post: Literal["list"],
) -> Mapping[UH, list[T]]: ...
@overload
def bucket_mapping[T, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: None = None,
    post: Literal["tuple"],
) -> Mapping[UH, tuple[T, ...]]: ...
@overload
def bucket_mapping[T, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: None = None,
    post: Literal["set"],
) -> Mapping[UH, set[T]]: ...
@overload
def bucket_mapping[T, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: None = None,
    post: Literal["frozenset"],
) -> Mapping[UH, frozenset[T]]: ...
@overload
def bucket_mapping[T, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: None = None,
    post: Literal["unique"],
) -> Mapping[UH, T]: ...
@overload
def bucket_mapping[T, U, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: Callable[[T], U] | None = None,
    post: None = None,
) -> Mapping[UH, Iterator[U]]: ...
@overload
def bucket_mapping[T, U, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: Callable[[T], U],
    post: Literal["list"],
) -> Mapping[UH, list[U]]: ...
@overload
def bucket_mapping[T, U, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: Callable[[T], U],
    post: Literal["tuple"],
) -> Mapping[UH, tuple[U, ...]]: ...
@overload
def bucket_mapping[T, U, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: Callable[[T], U],
    post: Literal["set"],
) -> Mapping[UH, set[U]]: ...
@overload
def bucket_mapping[T, U, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: Callable[[T], U],
    post: Literal["frozenset"],
) -> Mapping[UH, frozenset[U]]: ...
@overload
def bucket_mapping[T, U, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: Callable[[T], U] | None = None,
    post: Literal["unique"],
) -> Mapping[UH, U]: ...
@overload
def bucket_mapping[T, U, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: Callable[[T], U] | None = None,
    post: Literal["list", "tuple", "set", "frozenset", "unique"] | None = None,
) -> (
    Mapping[UH, Iterator[T]]
    | Mapping[UH, list[T]]
    | Mapping[UH, tuple[T, ...]]
    | Mapping[UH, set[T]]
    | Mapping[UH, frozenset[T]]
    | Mapping[UH, T]
    | Mapping[UH, Iterator[U]]
    | Mapping[UH, list[U]]
    | Mapping[UH, tuple[U, ...]]
    | Mapping[UH, set[U]]
    | Mapping[UH, frozenset[U]]
    | Mapping[UH, U]
): ...
def bucket_mapping[T, U, UH: Hashable](
    iterable: Iterable[T],
    func: Callable[[T], UH],
    /,
    *,
    pre: Callable[[T], U] | None = None,
    post: Literal["list", "tuple", "set", "frozenset", "unique"] | None = None,
) -> (
    Mapping[UH, Iterator[T]]
    | Mapping[UH, list[T]]
    | Mapping[UH, tuple[T, ...]]
    | Mapping[UH, set[T]]
    | Mapping[UH, frozenset[T]]
    | Mapping[UH, T]
    | Mapping[UH, Iterator[U]]
    | Mapping[UH, list[U]]
    | Mapping[UH, tuple[U, ...]]
    | Mapping[UH, set[U]]
    | Mapping[UH, frozenset[U]]
    | Mapping[UH, U]
):
    """Bucket the values of iterable into a mapping."""
    bckt = bucket(iterable, func)
    mapping = {key: bckt[key] for key in bckt}
    match pre, post:
        case None, None:
            return mapping
        case None, "list":
            return {k: list(v) for k, v in mapping.items()}
        case None, "tuple":
            return {k: tuple(v) for k, v in mapping.items()}
        case None, "set":
            return {k: set(v) for k, v in mapping.items()}
        case None, "frozenset":
            return {k: frozenset(v) for k, v in mapping.items()}
        case None, "unique":
            return _bucket_mapping_unique(mapping)
        case Callable(), None:
            return {k: map(pre, v) for k, v in mapping.items()}
        case Callable(), "list":
            return {k: list(map(pre, v)) for k, v in mapping.items()}
        case Callable(), "tuple":
            return {k: tuple(map(pre, v)) for k, v in mapping.items()}
        case Callable(), "set":
            return {k: set(map(pre, v)) for k, v in mapping.items()}
        case Callable(), "frozenset":
            return {k: frozenset(map(pre, v)) for k, v in mapping.items()}
        case Callable(), "unique":
            return _bucket_mapping_unique({k: map(pre, v) for k, v in mapping.items()})
        case _ as never:
            assert_never(never)


def _bucket_mapping_unique[K: Hashable, V](
    mapping: Mapping[K, Iterable[V]], /
) -> Mapping[K, V]:
    results: dict[K, V] = {}
    errors: dict[K, tuple[V, V]] = {}
    for key, value in mapping.items():
        try:
            results[key] = one(value)
        except OneNonUniqueError as error:
            errors[key] = (error.first, error.second)
    if len(errors) >= 1:
        raise BucketMappingError(errors=errors)
    return results


@dataclass(kw_only=True, slots=True)
class BucketMappingError[K: Hashable, V](Exception):
    errors: Mapping[K, tuple[V, V]]

    @override
    def __str__(self) -> str:
        parts = [
            f"{get_repr(key)} (#1: {get_repr(first)}, #2: {get_repr(second)})"
            for key, (first, second) in self.errors.items()
        ]
        desc = ", ".join(parts)
        return f"Buckets must contain exactly one item each; got {desc}"


##


def partition_list[T](
    pred: Callable[[T], bool], iterable: Iterable[T], /
) -> tuple[list[T], list[T]]:
    """Partition with lists."""
    false, true = partition(pred, iterable)
    return list(false), list(true)


##


def partition_typeguard[T, U](
    pred: Callable[[T], TypeGuard[U]], iterable: Iterable[T], /
) -> tuple[Iterator[T], Iterator[U]]:
    """Partition with a typeguarded function."""
    false, true = partition(pred, iterable)
    true = cast("Iterator[U]", true)
    return false, true


##


class peekable[T](_peekable):  # noqa: N801
    """Peekable which supports dropwhile/takewhile methods."""

    def __init__(self, iterable: Iterable[T], /) -> None:
        super().__init__(iterable)

    @override
    def __iter__(self) -> Iterator[T]:  # pyright: ignore[reportIncompatibleMethodOverride]
        while bool(self):
            yield next(self)

    @override
    def __next__(self) -> T:
        return super().__next__()

    def dropwhile(self, predicate: Callable[[T], bool], /) -> None:
        while bool(self) and predicate(self.peek()):
            _ = next(self)

    @overload
    def peek(self, *, default: Sentinel = sentinel) -> T: ...
    @overload
    def peek[U](self, *, default: U) -> T | U: ...
    @override
    def peek(self, *, default: Any = sentinel) -> Any:  # pyright: ignore[reportIncompatibleMethodOverride]
        if isinstance(default, Sentinel):
            return super().peek()
        return super().peek(default=default)

    def takewhile(self, predicate: Callable[[T], bool], /) -> Iterator[T]:
        while bool(self) and predicate(self.peek()):
            yield next(self)


##


@dataclass(kw_only=True, slots=True)
class Split[T]:
    """An iterable split into head/tail."""

    head: T
    tail: T

    @override
    def __repr__(self) -> str:
        cls = get_class_name(self)
        spaces = 4 * " "
        head_first = indent("head=", spaces)
        head_rest = indent(repr(self.head), 2 * spaces)
        tail_first = indent("tail=", spaces)
        tail_rest = indent(repr(self.tail), 2 * spaces)
        joined = f"{head_first}\n{head_rest}\n{tail_first}\n{tail_rest}"
        return f"{cls}(\n{joined}\n)"


def yield_splits[T](
    iterable: Iterable[T],
    head: int,
    tail: int,
    /,
    *,
    min_frac: float | None = None,
    freq: int | None = None,
) -> Iterator[Split[Sequence[T]]]:
    """Yield the splits of an iterable."""
    it1 = _yield_splits1(iterable, head + tail)
    it2 = _yield_splits2(it1, head, tail, min_frac=min_frac)
    it3 = _yield_splits3(it2)
    freq_use = tail if freq is None else freq
    return islice(it3, 0, None, freq_use)


def _yield_splits1[T](
    iterable: Iterable[T], total: int, /
) -> Iterator[tuple[Literal["head", "body"], Sequence[T]]]:
    peek = peekable(iterable)
    for i in range(1, total + 1):
        if len(result := peek[:i]) < i:
            return
        yield "head", result
    while True:
        _ = next(peek)
        if len(result := peek[:total]) >= 1:
            yield "body", result
        else:
            break


def _yield_splits2[T](
    iterable: Iterable[tuple[Literal["head", "body"], Sequence[T]],],
    head: int,
    tail: int,
    /,
    *,
    min_frac: float | None = None,
) -> Iterator[tuple[Iterable[T], int, int]]:
    min_length = head if min_frac is None else min_frac * head
    for kind, window in iterable:
        len_win = len(window)
        match kind:
            case "head":
                len_head = max(len_win - tail, 0)
                if len_head >= min_length:
                    yield window, len_head, tail
            case "body":
                len_tail = max(len_win - head, 0)
                if len_tail >= 1:
                    yield window, head, len_tail
            case _ as never:
                assert_never(never)


def _yield_splits3[T](
    iterable: Iterable[tuple[Iterable[T], int, int]], /
) -> Iterator[Split[Sequence[T]]]:
    for window, len_head, len_tail in iterable:
        head_win, tail_win = split_into(window, [len_head, len_tail])
        yield cast(
            "Split[Sequence[T]]", Split(head=list(head_win), tail=list(tail_win))
        )


__all__ = [
    "BucketMappingError",
    "Split",
    "bucket_mapping",
    "partition_list",
    "partition_typeguard",
    "peekable",
    "yield_splits",
]



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from ipaddress import IPv4Address, IPv6Address
from os import environ
from pathlib import Path
from re import search
from typing import TYPE_CHECKING, Any, assert_never, override

import typed_settings
from typed_settings import EnvLoader, FileLoader, find
from typed_settings.converters import TSConverter
from typed_settings.loaders import TomlFormat
from whenever import (
    Date,
    DateDelta,
    DateTimeDelta,
    MonthDay,
    PlainDateTime,
    Time,
    TimeDelta,
    YearMonth,
    ZonedDateTime,
)

from utilities.iterables import always_iterable
from utilities.pathlib import get_path
from utilities.string import substitute_environ

if TYPE_CHECKING:
    from collections.abc import Callable

    from typed_settings.loaders import Loader
    from typed_settings.processors import Processor

    from utilities.types import MaybeCallablePathLike, MaybeIterable, PathLike


##


class ExtendedTSConverter(TSConverter):
    """An extension of the TSConverter for custom types."""

    @override
    def __init__(
        self,
        *,
        resolve_paths: bool = True,
        strlist_sep: str | Callable[[str], list] | None = ":",
    ) -> None:
        super().__init__(resolve_paths=resolve_paths, strlist_sep=strlist_sep)
        cases: list[tuple[type[Any], Callable[..., Any]]] = [
            (Date, Date.parse_common_iso),
            (DateDelta, DateDelta.parse_common_iso),
            (DateTimeDelta, DateTimeDelta.parse_common_iso),
            (IPv4Address, IPv4Address),
            (IPv6Address, IPv6Address),
            (MonthDay, MonthDay.parse_common_iso),
            (Path, partial(_parse_path, resolve=resolve_paths, pwd=Path.cwd())),
            (PlainDateTime, PlainDateTime.parse_common_iso),
            (Time, Time.parse_common_iso),
            (TimeDelta, TimeDelta.parse_common_iso),
            (YearMonth, YearMonth.parse_common_iso),
            (ZonedDateTime, ZonedDateTime.parse_common_iso),
        ]
        extras = {cls: _make_converter(cls, func) for cls, func in cases}
        self.scalar_converters |= extras


def _make_converter[T](
    cls: type[T], parser: Callable[[str], T], /
) -> Callable[[Any, type[Any]], Any]:
    def hook(value: T | str, _: type[T] = cls, /) -> Any:
        if not isinstance(value, (cls, str)):  # pragma: no cover
            msg = f"Invalid type {type(value).__name__!r}; expected '{cls.__name__}' or 'str'"
            raise TypeError(msg)
        if isinstance(value, str):
            return parser(value)
        return value

    return hook


def _parse_path(
    path: str, /, *, resolve: bool = False, pwd: MaybeCallablePathLike | None = None
) -> Path:
    path = substitute_environ(path, **environ)
    match resolve:
        case True:
            return get_path(path=pwd).joinpath(path).resolve()
        case False:
            return Path(path)
        case _ as never:
            assert_never(never)


##


_BASE_DIR: Path = Path()


def load_settings[T](
    cls: type[T],
    app_name: str,
    /,
    *,
    filenames: MaybeIterable[str] = "settings.toml",
    start_dir: PathLike | None = None,
    loaders: MaybeIterable[Loader] | None = None,
    processors: MaybeIterable[Processor] = (),
    base_dir: Path = _BASE_DIR,
) -> T:
    if not search(r"^[A-Za-z]+(?:_[A-Za-z]+)*$", app_name):
        raise LoadSettingsError(appname=app_name)
    filenames_use = list(always_iterable(filenames))
    start_dir_use = None if start_dir is None else Path(start_dir)
    files = [find(filename, start_dir=start_dir_use) for filename in filenames_use]
    file_loader = FileLoader(formats={"*.toml": TomlFormat(app_name)}, files=files)
    env_loader = EnvLoader(f"{app_name.upper()}__", nested_delimiter="__")
    loaders_use: list[Loader] = [file_loader, env_loader]
    if loaders is not None:
        loaders_use.extend(always_iterable(loaders))
    return typed_settings.load_settings(
        cls,
        loaders_use,
        processors=list(always_iterable(processors)),
        converter=ExtendedTSConverter(),
        base_dir=base_dir,
    )


@dataclass(kw_only=True, slots=True)
class LoadSettingsError(Exception):
    appname: str

    @override
    def __str__(self) -> str:
        return f"Invalid app name; got {self.appname!r}"


__all__ = ["ExtendedTSConverter", "LoadSettingsError", "load_settings"]
